""" Sphinx extension to convert RST pages to notebooks """

from os import makedirs
from os.path import join as pjoin, dirname, isdir
from copy import deepcopy
from collections import defaultdict

from docutils import nodes, utils
from docutils.parsers.rst.roles import set_classes

from sphinx.util.nodes import split_explicit_title, set_role_source_info
from sphinx.errors import ExtensionError

# Use notebook format version 4
from .ipython_shim import nbf, nbconvert as nbc

from . import doctree2nb, doctree2py
from .sphinx2foos import PythonBuilder, NotebookBuilder
from .converters import UnicodeOutput


class RunRoleError(ExtensionError):
    """ Error for runnable role Sphinx extensions """


class runrole_reference(nodes.reference):
    """Node for references to built runnable, similar to pending_xref."""


class PyRunRole(object):
    """ Role builder for Python code file """

    default_text = 'Download this page as a Python code file'
    default_extension = '.py'
    code_type = 'pyfile'
    builder_class = PythonBuilder
    encoding = 'utf-8'

    _cache = {}

    def __call__(self, name, rawtext, text, lineno, inliner, options={},
                 content=[]):
        """ Register this document for later building as runnable code file.

        Parameters
        ----------
        name : str
            The role name used in the document.
        rawtext : str
            The entire markup snippet, including the role markup.
        text : str
            The text marked with the role.
        lineno : int
            The line number where `rawtext` appears in the input.
        inliner : object
            The inliner instance that called us.
        options : dict, optional
            Directive options for customization.
        content : content, optional
            The directive content for customization.

        Returns
        -------
        nodes : list
            list of nodes to insert into the document. Can be empty.
        messages : list
            list of system messages. Can be empty.
        """
        # Get objects from context
        env = inliner.document.settings.env
        # process class options.
        # http://docutils.sourceforge.net/docs/howto/rst-roles.html
        # Remaining options will be added as attributes of the node (see
        # below).
        set_classes(options)
        # Get title and link
        text = utils.unescape(text)
        text = self.default_text if text.strip() == '.' else text
        has_fname, title, fname = split_explicit_title(text)
        if not has_fname:
            fname = '/' + env.docname + self.default_extension
        refnode = runrole_reference(rawtext, title,
                                    reftype=self.code_type,
                                    refdoc=env.docname,
                                    reftarget=fname,
                                    **options)
        # We may need the line number for warnings
        set_role_source_info(inliner, lineno, refnode)
        return [refnode], []

    def write_queue(self, queue, app):
        """ Write queue of runnable nodes

        Parameters
        ----------
        queue : iterable
            Iterable of Docutils nodes, where the nodes specify runnable
            builds, including (for each node) the filename of original ReST
            document.
        app : Sphinx Application
            Application responsible for build.
        """
        for node in queue:
            self.write(node, app)

    def write(self, node, app):
        """ Build + cache runnable, or return build from cache.

        Parameters
        ----------
        node : docutils node
            Docutils node specifying runnable build, including filename of
            original ReST document.
        app : Sphinx Application
            Application responsible for build.
        """
        out_fname = _relfn2outpath(node['filename'], app)
        built = self.get_built(node, app)
        path = dirname(out_fname)
        if not isdir(path):
            makedirs(path)
        with open(out_fname, 'wb') as fobj:
            fobj.write(built.encode(self.encoding))

    def get_built(self, node, app):
        """ Build, cache, return output object, or return from cache.

        Parameters
        ----------
        node : object
            Runrole node.
        app : object
            Sphinx application in charge of build

        Returns
        -------
        output : built output.
        """
        code_type = self.code_type
        own_params = app.env.runrole_cache[node['refdoc']]
        if own_params.get(code_type) is None:
            own_params[code_type] = self._build(node, app)
        return own_params[code_type]

    def _build(self, node, app):
        """ Return string containing built / resolved version of `doctree`
        """
        builder = self.builder_class(app)
        docname = node['refdoc']
        doctree = app.env.get_and_resolve_doctree(docname, builder)
        builder.prepare_writing([docname])
        # Set current docname for writer to work out link targets
        builder.current_docname = docname
        return builder.writer.write(doctree, UnicodeOutput())


class ClearNotebookRunRole(PyRunRole):
    """ Role builder for not-evaluated notebook """

    default_text = 'Download this page as a Jupyter notebook (no outputs)'
    default_extension = '.ipynb'
    code_type = 'clearnotebook'
    builder_class = NotebookBuilder


def convert_timeout(argument):
    """ Allow -1, 0, positive integers and None

    These are the valid options for the nbconvert timeout option.
    """
    if argument.lower() == 'none':
        return None
    value = int(argument)
    if value < -1:
        raise ValueError('Value less than -1 not allowed')
    return value


class FullNotebookRunRole(ClearNotebookRunRole):
    """ Role builder for evaluated notebook """

    default_text = 'Download this page as a Jupyter notebook (with outputs)'
    code_type = 'fullnotebook'

    options = dict(timeout=convert_timeout)

    def __init__(self, clear_role):
        self.clear_role = clear_role

    def write_queue(self, queue, app):
        """ Write queue of runnable nodes

        Choose longest timeout for duplicate builds.

        Parameters
        ----------
        queue : iterable
            Iterable of Docutils nodes, where the nodes specify runnable
            builds, including (for each node) the filename of original ReST
            document.
        app : Sphinx Application
            Application responsible for build.
        """
        # Set longest timeout to duplicates
        for docname in set(n['refdoc'] for n in queue):
            duplicates = [n for n in queue if n['refdoc'] == docname]
            timeouts = set(n['timeout'] for n in duplicates if 'timeout' in n)
            if len(timeouts) < 2:
                continue
            max_timeout = (-1 if {None, -1}.intersection(timeouts)
                           else max(timeouts))
            for n in duplicates:
                n['timeout'] = max_timeout

        for node in queue:
            self.write(node, app)

    def _build(self, node, env):
        """ Return byte string containing built version of `doctree` """
        empty_json = self.clear_role.get_built(node, env)
        timeout = node.get('timeout', env.config.fill_notebook_timeout)
        full_nb = fill_notebook(nbf.reads(empty_json), timeout=timeout)
        return nbf.writes(full_nb)


# Collect instances of the known role types
_clearnotebook = ClearNotebookRunRole()
NAME2ROLE = dict(pyfile=PyRunRole(),
                 codefile=PyRunRole(),  # compatibility with older name
                 clearnotebook=_clearnotebook,
                 fullnotebook=FullNotebookRunRole(_clearnotebook))


def do_builder_init(app):
    """ Initialize builder with empty runrole caches and queues.
    """
    env = app.env
    env.runrole_queue = defaultdict(list)
    env.runrole_cache = defaultdict(dict)


def do_purge_doc(app, env, docname):
    """ Clear caches and queues of runrole builds from this `docname`.
    """
    env.runrole_cache.pop(docname, None)
    queues = env.runrole_queue
    for code_type in queues:
        queues[code_type] = [node for node in queues[code_type]
                             if node['refdoc'] != docname]


def collect_runfiles(app, doctree, fromdocname):
    r""" Collate requested runnable files, store in env

    Traverse doctree, find runrole nodes, and collect the nodes that need to be
    built.  Store references to the nodes in the ``env.runroles_queue``
    dictionary, The dictionary has keys giving runrole type (string, one
    of 'pyfile', 'clearnotebook', 'fullnotebook') and values that are lists
    of nodes to be built into runnable outputs.

    Set filename of file to be built into node.

    Called at ``doctree-resolved`` event.
    """
    env = app.env
    queues = env.runrole_queue
    files = {}

    for ref in doctree.traverse(runrole_reference):
        # Calculate filename of file to be built, relative to project root
        rel_fn, _ = env.relfn2path(ref['reftarget'], ref['refdoc'])
        ref['filename'] = rel_fn
        # Check for duplicates.  It's OK to reference a file that is already
        # registered for building, but it must be of the same code type.
        code_type = ref['reftype']
        if rel_fn not in files:
            files[rel_fn] = code_type
        elif files[rel_fn] != code_type:
            raise RunRoleError(
                'Trying to register filename {0} as type {1}, '
                'but it is already registered as type {2}'.format(
                    rel_fn,
                    code_type,
                    files[rel_fn]))
        # The queue can have duplicate combinations of (docname, code_type,
        # rel_fn).
        queues[code_type].append(ref)


def _relfn2outpath(rel_path, app):
    return pjoin(app.outdir, rel_path)


def write_runfiles(app, exception):
    """ Write notebooks / code files when build has finished

    :func:`collect_runfiles` has already collected the files that need to be
    built, and stored then in the ``env.runroles`` dictionary.  See the
    docstring for that function for details.

    We cycle through these collected relative filenames, and build the
    necessary files using the ``write`` method of the stored role instances.

    Called at ``build-finished`` event.
    """
    if exception is not None:
        return
    for code_type, queue in app.env.runrole_queue.items():
        NAME2ROLE[code_type].write_queue(queue, app)


def visit_runrole(self, node):
    self.body.append(
        '<a class="reference download internal" href="{0}">'.format(
            node['filename']))
    self.context.append('</a>')


def depart_runrole(self, node):
    self.body.append(self.context.pop())


def fill_notebook(nb, timeout=30):
    """ Execute notebook `nb` and return notebook with built outputs
    """
    preprocessor = nbc.preprocessors.execute.ExecutePreprocessor(
        timeout=timeout)
    preprocessor.enabled = True
    res = nbc.exporter.ResourcesDict()
    res['metadata'] = nbc.exporter.ResourcesDict()
    output_nb, _ = preprocessor(deepcopy(nb), res)
    return output_nb


def drop_visit(self, node):
    raise nodes.SkipNode


def setup(app):
    # Add runrole roles
    for name, role in NAME2ROLE.items():
        app.add_role(name, role)
    # Create dictionaries in builder environment
    app.connect(str('builder-inited'), do_builder_init)
    # Delete caches when document re-initialized
    app.connect('env-purge-doc', do_purge_doc)
    # Collect and check all runrole nodes when doctree done
    app.connect('doctree-resolved', collect_runfiles)
    # Write output files at end of build
    app.connect('build-finished', write_runfiles)
    app.add_node(runrole_reference,
                 html=(visit_runrole, depart_runrole),
                 text=(drop_visit, None),
                 latex=(drop_visit, None))
    # Register translators to allow other extensions to extend ipynb and python
    # code visit, depart methods with app.add_node as we have just done for the
    # html translator in the lines above.  See:
    # http://www.sphinx-doc.org/en/1.4.8/extdev/tutorial.html#the-setup-function
    app.set_translator('pyfile', doctree2py.Translator)
    app.set_translator('ipynb', doctree2nb.Translator)



""" Test whether doc examples same as previously
"""

from os.path import join as pjoin, dirname

from nb2plots.from_notebook import convert_nb_fname
from nb2plots.converters import to_py, to_notebook

from nb2plots.testing.convutils import fcontents
from nb2plots.testing.nbtesters import assert_nb_equiv

HERE = dirname(__file__)
DATA = pjoin(HERE, 'data')


def test_regression():
    # Test documentation worked example
    input_nb_fname = pjoin(DATA, 'example_notebook.ipynb')
    output_rst_fname = pjoin(DATA, 'converted_example.rst')
    # Convert to ReST, add trailing CR from output script
    rst = convert_nb_fname(input_nb_fname) + '\n'
    assert rst.encode('utf8') == fcontents(output_rst_fname)
    # Convert ReST to output formats
    py_file = to_py.from_rst(rst)
    assert (py_file.encode('utf8') ==
            fcontents(pjoin(DATA, 'converted_plus_notebooks.py')))
    ipy_file = to_notebook.from_rst(rst)
    assert_nb_equiv(ipy_file,
                 fcontents(pjoin(DATA, 'converted_plus_notebooks.ipynb'))
                    .decode('utf8'))



""" Tests for Markdown and other builders
"""
from __future__ import unicode_literals

from os.path import join as pjoin, isfile

from nb2plots.testing import PlotsBuilder


class TestMarkdownBuild(PlotsBuilder):
    """ Markdown builder without specified base URL
    """

    builder = 'markdown'

    rst_sources = {'a_page': """\
.. _a-ref:

################
Refereed section
################

This section refers to :ref:`itself <a-ref>`.

It also refers forward to the :ref:`next section <b-ref>`.

Then, and finally, it refers to itself with its own name: :ref:`a-ref`.

.. _b-ref:

##########
Rerefereed
##########

This section refers to this document at :doc:`a_page`, and with an
explicit title, to :doc:`this document <a_page>`.

Then to :doc:`a_page`.  Again to :doc:`another doc <a_page>`.

Now :download:`a_page.rst`.

Then :download:`another page <a_page.rst>`.

.. code-links::

Then `a link <https://another-place.com/page.html>`_.

Again, we :doc:`link to another doc <subdir1/b_page>`.""",
                   'subdir1/b_page': """\
############
Another page
############

Here is another page.

It refers to :doc:`/a_page`.""",
                   'subdir2/c_page': """\
############
Further page
############

Here is further page.

It refers to :doc:`../a_page`.

It also refers to :doc:`/subdir1/b_page`."""}

    def test_output(self):
        assert self.get_built_file('contents.md').strip() == ''
        assert self.get_built_file('a_page.md') == """\
## Refereed section

This section refers to itself.

It also refers forward to the next section.

Then, and finally, it refers to itself with its own name: Refereed section.

## Rerefereed

This section refers to this document at Refereed section, and with an
explicit title, to this document.

Then to Refereed section.  Again to another doc.

Now `a_page.rst`.

Then `another page`.

Then [a link](https://another-place.com/page.html).

Again, we link to another doc.
"""
        assert self.get_built_file(pjoin('subdir1', 'b_page.md')) == """\
## Another page

Here is another page.

It refers to Refereed section.
"""
        assert self.get_built_file(pjoin('subdir2', 'c_page.md')) == """\
## Further page

Here is further page.

It refers to Refereed section.

It also refers to Another page.
"""


class TestBasedMarkdownBuild(TestMarkdownBuild):
    """ Markdown builder with specified base URL
    """

    conf_source = ('extensions = ["nb2plots"]\n'
                   'markdown_http_base = "https://dynevor.org"')

    def test_output(self):
        assert self.get_built_file('contents.md').strip() == ''
        assert self.get_built_file('a_page.md') == """\
## Refereed section

This section refers to [itself](https://dynevor.org/a_page.html#a-ref).

It also refers forward to the [next section](https://dynevor.org/a_page.html#b-ref).

Then, and finally, it refers to itself with its own name: [Refereed section](https://dynevor.org/a_page.html#a-ref).

## Rerefereed

This section refers to this document at [Refereed section](https://dynevor.org/a_page.html), and with an
explicit title, to [this document](https://dynevor.org/a_page.html).

Then to [Refereed section](https://dynevor.org/a_page.html).  Again to [another doc](https://dynevor.org/a_page.html).

Now [a_page.rst](https://dynevor.org/_downloads/a_page.rst).

Then [another page](https://dynevor.org/_downloads/a_page.rst).

Then [a link](https://another-place.com/page.html).

Again, we [link to another doc](https://dynevor.org/subdir1/b_page.html).
"""
        assert self.get_built_file(pjoin('subdir1', 'b_page.md')) == """\
## Another page

Here is another page.

It refers to [Refereed section](https://dynevor.org/a_page.html).
"""
        assert self.get_built_file(pjoin('subdir2', 'c_page.md')) == """\
## Further page

Here is further page.

It refers to [Refereed section](https://dynevor.org/a_page.html).

It also refers to [Another page](https://dynevor.org/subdir1/b_page.html).
"""


class TestPythonBuild(PlotsBuilder):
    """ Python builder without specified base URL
    """

    builder = 'python'

    rst_sources = {'a_page': """\
.. _a-ref:

A section
=========

Some text

This section refers to :ref:`itself <a-ref>`.

.. nbplot::

    >>> a = 1
"""}

    def test_output(self):
        assert self.get_built_file('contents.py').strip() == ''
        assert self.get_built_file('a_page.py') == """\
# ## A section
#
# Some text
#
# This section refers to itself.

a = 1
"""


class TestBasedPythonBuild(TestPythonBuild):
    """ Python builder with specified base URL
    """

    conf_source = ('extensions = ["nb2plots"]\n'
                   'markdown_http_base = "https://dynevor.org"')

    def test_output(self):
        assert self.get_built_file('contents.py').strip() == ''
        assert self.get_built_file('a_page.py') == """\
# ## A section
#
# Some text
#
# This section refers to [itself](https://dynevor.org/a_page.html#a-ref).

a = 1
"""


class TestLatexBuild(PlotsBuilder):
    """ Test LaTeX build

    In particular, test that code outputs in a subdirectory work without
    raising an error.
    """

    builder = 'latex'

    rst_sources = {'foo/a_page': """\
A section
=========

.. code-links::
"""}

    def test_output(self):
        for suffix in ('.py', '.ipynb', '_full.ipynb'):
            assert isfile(pjoin(self.out_dir, 'foo', 'a_page' + suffix))



""" Shim for IPython before and after the big split
"""
try:
    import traitlets
    import traitlets.config as config
except ImportError:
    from IPython.utils import traitlets
    from IPython import config
try:
    import nbformat
except ImportError:
    from IPython import nbformat
try:
    import nbconvert
except ImportError:
    from IPython import nbconvert

# Use notebook format version 4 by default
nbf = nbformat.v4



""" Test strdiff module
"""

from nb2plots.strdiff import get_diff


def test_get_diff():
    assert get_diff('foo', 'foo') == "L0, OK: foo"
    assert get_diff('foo', 'baz') == "L0, S1: foo\nL0, S2: baz"
    assert (get_diff('foo\nfoo', 'bar\nfoo\nbaz') == """\
L0, S1: foo
L0, S2: bar
L1, OK: foo
Remaining line in second string is:
        baz""")
    assert (get_diff('bar\nfoo\nbaz', 'foo\nfoo') == """\
L0, S1: bar
L0, S2: foo
L1, OK: foo
Remaining line in first string is:
        baz""")
    assert (get_diff('bar\nfoo\nbaz\nbof', 'foo\nfoo') == """\
L0, S1: bar
L0, S2: foo
L1, OK: foo
Remaining lines in first string are:
        baz
        bof""")



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Test writer conversions

Test running writer over example files and chosen snippets
"""
from __future__ import (division, print_function, absolute_import,
                        unicode_literals)

import sys
from os.path import join as pjoin
from glob import glob
from functools import partial
from io import TextIOWrapper, BytesIO

from docutils import nodes
from docutils.utils import new_document
from docutils.core import publish_string

from nb2plots.doctree2md import Writer, IndentLevel

from nb2plots.testing import DATA_PATH
from nb2plots.testing.convutils import (convert_assert, doctree_assert,
                                        fcontents)


def assert_conv_equal(rst_str, md_expected, encoding='utf8'):
    converter = partial(publish_string, writer=Writer())
    convert_assert(rst_str, converter, md_expected, encoding)


def assert_dt_equal(doctree, md_expected):
    doctree_assert(doctree, Writer(), md_expected)


def test_example_files():
    # test rst2md script over all .rst files checking against .md files
    for rst_fname in glob(pjoin(DATA_PATH, '*.rst')):
        rst_contents = fcontents(rst_fname)
        md_fname = rst_fname[:-3] + 'md'
        md_contents = fcontents(md_fname)
        # Skip files containing text "skip".  These are files for which the
        # source ReST is not valid in plain docutils, such as those containing
        # Sphinx directives and roles.
        if md_contents.strip() != b'skip':
            assert_conv_equal(rst_contents, md_contents)


def test_indent_level():
    # Test IndentLevel object
    level = IndentLevel(['foo', 'bar'], 'prefix')
    assert len(level) == 0
    level.append('baz')
    assert len(level) == 1


def test_container():
    # Test container node passed through
    doc = new_document('md-test')
    container = nodes.container()
    doc.append(container)
    container.append(nodes.Text('Boo!'))
    assert_dt_equal(doc, b'Boo!\n')


def test_snippets():
    assert_conv_equal("Some text", b"Some text\n")
    assert_conv_equal("With *emphasis*", b"With *emphasis*\n")
    assert_conv_equal("That's **strong**", b"That's **strong**\n")
    assert_conv_equal("As ``literal``", b"As `literal`\n")
    assert_conv_equal("To ``defrole``", b"To `defrole`\n")
    assert_conv_equal("Now :math:`a = 1`", b"Now $a = 1$\n")


def test_system_message():
    # This output could surely be improved.
    old_stderr = sys.stderr
    sys.stderr = TextIOWrapper(BytesIO())
    try:
        assert_conv_equal('Text. :bad-role:`.`.  More text.', b"""\
Text. 

```
:bad-role:`.`
```

.  More text.

```
System Message: <string>:, line 1

<string>:1: (ERROR/3) Unknown interpreted text role "bad-role".
```
""")
    finally:
        sys.stderr = old_stderr



""" Mocking Sphinx apps for testing
"""

try:
    # Python >=3.3
    from unittest import mock
except ImportError:
    import mock

from sphinx.application import Sphinx
from sphinx.config import Config


def get_app():
    app = mock.Mock(Sphinx)
    app.confdir = None
    app.config = mock.Mock(Config)
    return app



""" Convert doctree to Jupyter notebook
"""
from __future__ import unicode_literals

from docutils import nodes

from .ipython_shim import nbf
from . import doctree2py as d2py


class Translator(d2py.Translator):

    def _init_output(self):
        self._notebook = nbf.new_notebook()

    def _add_text_block(self, txt):
        self._notebook['cells'].append(nbf.new_markdown_cell(txt))

    def astext(self):
        """ Return the document as a string """
        self.flush_text()
        return nbf.writes(self._notebook)

    def add_code_block(self, txt):
        self.flush_text()
        self._notebook['cells'].append(nbf.new_code_cell(txt))

    def visit_mpl_hint(self, node):
        self.add_code_block('%matplotlib inline')
        raise nodes.SkipNode


class Writer(d2py.Writer):
    supported = ('jupyter',)
    """Formats this writer supports."""

    translator_class = Translator



# ## Docinfo Example
#
# author
#
# :   Our noble selves
#
# contact
#
# :   [ons@organization.com](mailto:ons@organization.com)
#
# The items above become "docinfo" in Docutils, and "field"s in Sphinx.
#
# author
#
# :   J Random Hacker
#
# contact
#
# :   [jrh@example.com](mailto:jrh@example.com)
#
# date
#
# :   2002-08-18
#
# status
#
# :   Work In Progress
#
# version
#
# :   1
#
# filename
#
# :   $RCSfile$
#
# copyright
#
# :   This document has been placed in the public domain.  You are
#     welcome to claim this as your own work.  It can go into your
#     thesis.   You can publish it in a book.  You can patent it.  You
#     can do anything you like with it.
#
# The above are "field"s in Docutils and Sphinx.
